//==========================================================================
//	File:		Monitor.cs
//
//	Summary:	Example shows use of Monitor in threads.
//					Monitor.Enter(Object)
//					Monitor.Exit(Object)
//					Monitor.Pulse(Object)
//					Monitor.Wait(Object)
//
//----------------------------------------------------------------------------
//	This file is part of the Microsoft NGWS Runtime Samples.
//
//	Copyright (C) 1998-2000 Microsoft Corporation.  All rights reserved.
//===========================================================================
using System;
using System.Threading;

public class MonitorSample
{
	public static void Main(String[] args)
	{
		int			result		= 0;		//	result initialized to say there is no error
		Cell		cell		= new Cell( );

		CellProd	prod		= new CellProd(cell,20);	//	use cell for storage, produce 20 items
		CellCons	cons		= new CellCons(cell,20);	//	use cell for storage, consume 20 items

		Thread		producer	= new Thread(new ThreadStart(prod.ThreadRun));
		Thread		consumer	= new Thread(new ThreadStart(cons.ThreadRun));
		//	threads producer and consumer have been created, but not started at this point.

		try
		{
			producer.Start( );
			consumer.Start( );

			producer.Join( );		//	Join with no Stop.  Run producer thread until it is done.
			consumer.Join( );		//	Join with no Stop.  Run consumer thread until it is done.
		//	threads producer and consumer are dead at this point.
		}
		catch (ThreadStateException e)
		{
			Console.WriteLine(e);			//	display text of exception
			result = 1;						//	result says there was an error
		}
		catch (ThreadInterruptedException e)
		{
			Console.WriteLine(e);			//	display text of exception
			result = 1;						//	result says there was an error
		}
		Environment.ExitCode = result;		//	eventhough Main with a return type of void, this sets
	}										//		a return code to parent process.
}	//	End of public class MonitorSample


public class CellProd
{
	Cell	cell;							//	Field to hold cell object to be used
	int		quantity	= 1;				//	Field for "How many items to produce in cell?"

	public CellProd(Cell box,int request)
	{
		cell = box;							//	Pass in what cell object to be used
		quantity = request;					//	Pass in "How many items to produce in cell?"
	}
	public void ThreadRun( )
	{
		for(int looper=1;looper<=quantity;)
			cell.WriteToCell(looper++);		//	"producing"
	}
}	//	End of public class CellProd

public class CellCons
{
	Cell	cell;							//	Field to hold cell object to be used
	int		quantity	= 1;				//	Field for "How many items to consume from cell?"

	public CellCons(Cell box,int request)
	{
		cell = box;							//	Pass in what cell object to be used
		quantity = request;					//	Pass in "How many items to consume from cell?"
	}
	public void ThreadRun( )
	{
		int		valReturned;
		for(int looper=1;looper<=quantity;looper++)
			valReturned=cell.ReadFromCell( );	//	if you wanted to "consume" the result, see valReturned
	}
}	//	End of public class CellCons


public class Cell
{
	int		cellContents;				//	Cell Contentents
	bool	readerFlag		= false;	//	State Flag
	public int ReadFromCell( )
	{
		Monitor.Enter(this);			//	Enter Synchronization block
		if (!readerFlag)
		{								//	Wait until Cell.WriteToCell is done doing its producing
			try
			{
				Monitor.Wait(this);		//	Waits until it gets a Monitor.Pulse(this)
			}
			catch (SynchronizationLockException e)
			{
				Console.WriteLine(e);
			}
			catch (ThreadInterruptedException e)
			{
				Console.WriteLine(e);
			}
		}
		Console.WriteLine("Consume: {0}",cellContents);
		readerFlag = false;			//	Reset the State Flag to say, done
		Monitor.Pulse(this);		//	Pulse tells Cell.WriteToCell that Cell.ReadFromCell is done.
		Monitor.Exit(this);			//	Exit Synchronization block
		return cellContents;
	}	//	End of method public int ReadFromCell( )
	
	public void WriteToCell(int n)
	{
		Monitor.Enter(this);			//	Enter Synchronization block
		if (readerFlag)
		{								//	Wait until Cell.ReadFromCell is done doing its consuming
			try
			{
				Monitor.Wait(this);		//	Waits until it gets a Monitor.Pulse(this)
			}
			catch (SynchronizationLockException e)
			{
				Console.WriteLine(e);
			}
			catch (ThreadInterruptedException e)
			{
				Console.WriteLine(e);
			}
		}
		cellContents = n;
		Console.WriteLine("Produce: {0}",cellContents);
		readerFlag = true;			//	Reset the State Flag to say, done
		Monitor.Pulse(this);		//	Pulse tells Cell.ReadFromCell that Cell.WriteToCell is done.
		Monitor.Exit(this);			//	Exit Synchronization block
	}	//	End of method public void WriteToCell(int value)
}	//	End of public class Cell

/*
Output

Produce: 1
Consume: 1
Produce: 2
Consume: 2
Produce: 3
Consume: 3
Produce: 4
Consume: 4
Produce: 5
Consume: 5
Produce: 6
Consume: 6
Produce: 7
Consume: 7
Produce: 8
Consume: 8
Produce: 9
Consume: 9
Produce: 10
Consume: 10
Produce: 11
Consume: 11
Produce: 12
Consume: 12
Produce: 13
Consume: 13
Produce: 14
Consume: 14
Produce: 15
Consume: 15
Produce: 16
Consume: 16
Produce: 17
Consume: 17
Produce: 18
Consume: 18
Produce: 19
Consume: 19
Produce: 20
Consume: 20

*/